// tsynth test

// changed to 7-byte string for serial data

/*
; Storm_I_FPGA pin assigns
;
; clk      93
; led1     96
; led2     94
; led3     91
; led4     85
; led5     84
; led6     83
; led7     82 
; led8     78
; key1     100
; key2     99
; key3     98
; key4     97
; reset    76
; sertx    70
; serrx    72
; beep     74
; ps2clk   52
; ps2data  50
*/

module stsynth (
input clk,
input serrx,
output reg ps2data,
output reg ps2clk,
output reg led1,
output reg led2,
output reg led3,
output reg led4,
output reg led6,
output reg led7,
output reg sertx
);

reg clk2;

reg reset;
reg [7:0] regaddr;
reg [15:0] regdatain;
reg rwrite;
wire signed [15:0] audleft;
wire signed [15:0] audright;
//wire [7:0] debug1;

wire signed [15:0] audleft2;
wire signed [15:0] audright2;

tsynth synth1 (clk, reset, regaddr, regdatain, rwrite, audleft2, audright2);

assign audleft = audleft2 >>> 2;                // lower amplitude to avoid clipping at output
assign audright = audright2 >>> 2;

initial begin
reset = 0;
rwrite = 0;
end

// ------------ audio output -------------

reg signed [16:0] sigdelt1;
reg signed [16:0] sigdelt2;

wire signed [16:0] higherr1 = 32767 - audleft;
wire signed [16:0] lowerr1 = audleft + 32768;
wire signed [16:0] higherr2 = 32767 - audright;
wire signed [16:0] lowerr2 = audright + 32768;

always @(posedge clk) begin

//led1 <= derp[23];
//led3 <= audleft[15];
//led4 <= ps2data;

if (sigdelt1<audleft) begin
        ps2data <= 1;
        sigdelt1 <= sigdelt1 + higherr1;
end else begin
        ps2data <= 0;
        sigdelt1 <= sigdelt1 - lowerr1;
        end

if (sigdelt2<audright) begin
        ps2clk <= 1;
        sigdelt2 <= sigdelt2 + higherr2;
end else begin
        ps2clk <= 0;
        sigdelt2 <= sigdelt2 - lowerr2;
        end

end


// ----------- SERIAL stuff -------------

reg [19:0] divider;

reg [19:0] rtimer;
reg [7:0] rshift;
reg [3:0] rbit;
reg [7:0] inbyte;
reg [3:0] debounce;
reg recvidle;
reg bytedone;
reg bytepresent;
reg byteack;

/*
reg [19:0] ttimer;
reg [7:0] tshift;
reg [3:0] tbit;
reg [7:0] outbyte;
reg sendbyte;
*/

initial divider = 434;          // 115200bps

initial recvidle = 1;
initial bytepresent = 0;

reg [2:0] regselect;
reg [3:0] nibble;
reg newnibble;

always @(posedge clk) begin

led2 <= regselect[0];
led3 <= regselect[1];
led4 <= regselect[2];

//led6 <= debug1[1];
//led7 <= debug1[0];

if (bytepresent) begin
        byteack <= 1;
        casex (inbyte)
        8'b01100001: regselect <= 1;                    // 'a' selects address register
//        8'b01100100: regselect <= 4;                    // 'd' selects data register
        8'b0011xxxx: begin
                newnibble <= 1;
                nibble <= inbyte[3:0];                  // digits 0-9
                end
        8'b01000xxx: begin
                newnibble <= 1;
                nibble <= inbyte[2:0] + 4'h9;           // digits A-F
                end
        endcase
end else begin
        byteack <= 0;
        end

if (newnibble) begin
        newnibble <= 0;
        regselect <= regselect + 1;
        case (regselect)
        3'b001: regaddr[7:4] <= nibble;
        3'b010: regaddr[3:0] <= nibble;
        3'b011: regdatain[15:12] <= nibble;
        3'b100: regdatain[11:8] <= nibble;
        3'b101: regdatain[7:4] <= nibble;
        3'b110: begin
                regdatain[3:0] <= nibble;
                rwrite <= 1;
                end
        endcase
        end

if (rwrite) rwrite <= 0;


// --------- serial data receiver ----------

if (serrx) begin
        if (debounce!=4'hF) debounce <= debounce + 1;
end else begin
        if (debounce!=4'h0) debounce <= debounce - 1;
        end

bytepresent <= bytedone & ~byteack;

if (recvidle) begin
        bytedone <= 0;
        if (debounce==4'h0) begin
                rtimer <= divider >> 1;
                rbit <= 0;
                recvidle <= 0;
                end
end else begin
        if (rtimer!=divider) begin
                rtimer <= rtimer + 1;
        end else begin
                rtimer <= 0;
                case (rbit)
                4'h0: begin
                        if (debounce==4'h0) rbit <= 1;
                                else recvidle <= 1;
                        end
                4'h9: begin
                        if (debounce==4'hF) begin
                                bytedone <= 1;
                                inbyte <= rshift;
                                end
                        recvidle <= 1;
                        end
                default: begin
                        rbit <= rbit + 1;
                        rshift <= {debounce[3], rshift[7:1]};
                        end
                endcase
                end
        end


// --------- serial data transmitter ----------

/*
if (ttimer!=divider) begin
        ttimer <= ttimer + 1;
end else begin
        case (tbit)
        4'h0: begin
                ttimer <= 0;
                tshift <= outbyte;                // data to send
                if (sendbyte) begin
                        tbit <= 1;
                        sertx <= 0;
                        end
                end
        4'h9: begin
//                ttimer <= divider >> 3;
                ttimer <= 0;
                tbit <= 0;
                sertx <= 1;
                sendbyte <= 0;
                end
        default: begin
                ttimer <= 0;
                tbit <= tbit + 1;
                sertx <= tshift[0];
                tshift <= tshift >> 1;
                end
        endcase
        end
*/

end




endmodule

