// parameter-based 'sound chip' by DamageX

// everything good!?

/*
16 operators per module, plus one noise generator

64 operator control registers
3 misc control registers
128 registers for user-defined waveforms
(sine, square, sawtooth, triangle waves are defined in internal ROM)

control registers (16 bit) for each operator:

$0x - panning and volume
        bits 15-11 = Right volume             (0=off, 31=max)
        bits 10-6 = Left volume               (0=off, 31=max)
        bits 5-0 = operator attenuation (0 = 0dB, 63 = -47.25dB)
$1x - octave and frequency
        bits 15-12 = octave
        bits 11-0 = frequency
$2x - waveform and modulation
        bit 15 = LFO select
        bit 14-12 = LFO sensitivity
        bit 11 = invert waveform
        bits 10-8 = wave pattern number
        bit 7 = noise enable
        bits 6-4 = self-modulation magnitude
        bits 3-2 = second modulator  \  0=self                   2=operator before that
        bits 1-0 = first modulator   /  1=previous operator      3=operator before THAT
$3x - envelope
        bits 15-12 = attack (0 disables envelope)
        bits 11-8 = sustain level (15 causes EG to skip right to decay2, use to do key-off)
        bits 7-4 = decay1
        bits 3-0 = decay2

EG has 5 states:
        000="DP0" resets bottom bits of EG value which are used to time DP1
        001="DP1" uses fixed time to release previous note
        01x=attack
        1x0=decay1
        1x1=decay2

misc register $00 - noise control
                        bits 6-4 = feedback bit
                        bits 2-0 = output bit
misc register $01 - noise period

misc register $02 - phase reset

misc register $03 - envelope time constant???

*/


module tsynth (clk, reset, regaddr, regdatain, rwrite, audleft, audright);

input clk;
input reset;
input [7:0] regaddr;
input [15:0] regdatain;
input rwrite;
output reg signed [15:0] audleft;
output reg signed [15:0] audright;
//output reg [7:0] debug1;


reg [15:0] regfile[0:63];

reg [15:0] waverom[0:255];
initial $readmemb("waveq64.txt", waverom, 0, 127);

// sine wave, square wave, sawtooth, triangle

reg [15:0] miscreg[0:3];

reg [15:0] resetp;                 // copy of miscreg[2] has key-on bits
reg resetthis;

reg [23:0] position[0:15];               // 6 address bits for wave pattern, + 18 fraction bits

reg signed [7:0] opvalue[0:15];          // last wave sample used by each channel

reg [20:0] chanleft;                     // result of 16-bit operator output * 5-bit panning
reg [20:0] chanright;                    // (...)
reg [19:0] leftsum;                      // sum of top 16 bits from each of 16 operators
reg [19:0] rightsum;                     // (...)

reg [2:0] clkcount;                      // counts from 0 to 4
reg [3:0] channel;                       // counts from 0 to 15
reg [3:0] oldchan;                       // lags channel by 5 cycles

reg [15:0] regread;                // data read from register file (once per clock)
reg [15:0] waveread;               // data read from waverom (once per clock, but not always used)
reg [7:0] waveaddr;                // address for waverom reads
reg [5:0] regaddr2;                // address for register reads
reg [7:0] regaddr3;                // address for register/waverom writes (latched from regaddr inputs)
reg [15:0] writedata;              // (latched from regdatain)
reg dowrite;                       // (latched from rwrite)

reg [23:0] chanpos;

reg [8:0] attenuator;              // multiplier for non-power-of-two volume levels
reg signed [15:0] tempvalue;
reg signed [15:0] tempvalue2;      // calculated sample for a channel (before panning)
reg signed [14:0] oldvalue2;       // copy of previous tempvalue2 (minus LSb)
reg signed [14:0] oldvalue3;       // copy of previous oldvalue3

reg [2:0] chanwave;
reg waveinvert;

reg [19:0] lfsr;                   // shift register for noise generator
reg [18:0] noisecount;
reg noisebit;

reg [2:0] chanfms;                 // LFO sensitivity
reg signed [7:0] chanfm;
reg signed [9:0] chanfm2;          // final pitch displacement

reg [9:0] chanpan;                 // left/right volume
reg [6:0] chanvol;                 // has an extra bit for envelope stuff

reg [3:0] chanoct;
reg [12:0] chanfreq;               // has an extra bit for LFO stuff

reg [3:0] chanpm;
reg signed [14:0] phasemod1;
reg signed [14:0] phasemod2;
reg [7:0] feedback;
reg [14:0] feedback2;              // feedback value with magnitude applied
reg highbyte;


// envelope generator stuff

reg [4:0] egcount;                 // used for global EG speed control
reg [2:0] egstage[0:15];
reg [2:0] chaneg;
reg [23:0] egvalue[0:15];          // EG value for each channel (6 bits + 18 fraction bits)
reg [23:0] egtemp;                 // working copy of EG value
reg [23:0] egtemp0;                // register that receives data from memory read (needed in Cyclone 1)
reg [3:0] ratetemp;
reg [3:0] egsustain;
reg egoverflow;
reg egstep;
reg egreset;
reg egrelease;


integer ii;
initial begin

// give things initial state to make simulator work
clkcount=0;       channel=0;         oldchan=0;
leftsum=0;        rightsum=0;
chanleft=0;       chanright=0;
tempvalue=0;      tempvalue2=0;
chanpan=0;        chanpos=0;
regread=0;        regaddr2=0;        waveaddr=0;

for (ii=0; ii<16; ii=ii+1) opvalue[ii]=0;
for (ii=0; ii<16; ii=ii+1) position[ii]=0;
for (ii=0; ii<16; ii=ii+1) egvalue[ii]=0;
for (ii=0; ii<64; ii=ii+1) regfile[ii]=0;
//for (ii=64; ii<128; ii=ii+1) waverom[ii]=0;             this was cancelling out the data from waveq.txt!?

// test data

miscreg[3]=16'h000D;

regfile[0]=16'h0000;
regfile[16]=16'h0040;

regfile[2]=16'h8400;
regfile[18]=16'h1800;
regfile[34]=16'h0000;

end


wire [7:0] chansamp0 = highbyte ? waveread[15:8] : waveread[7:0];
wire [7:0] chansamp = waveinvert ? ~chansamp0 : chansamp0;


always @(posedge clk) begin

casex (clkcount)                // cycles 0-4: read register data pertaining to channel
3'b000: begin
        if (channel==0) begin           // also output a sample clipped to 16 bits
                case (leftsum[19])
                1'b0: if (leftsum[18:15]==0) audleft <= leftsum[15:0];
                        else audleft <= 16'h7FFF;
                1'b1: if (leftsum[18:15]==4'b1111) audleft <= leftsum[15:0];
                        else audleft <= 16'h8000;
                endcase

                case (rightsum[19])
                1'b0: if (rightsum[18:15]==0) audright <= rightsum[15:0];
                        else audright <= 16'h7FFF;
                1'b1: if (rightsum[18:15]==4'b1111) audright <= rightsum[15:0];
                        else audright <= 16'h8000;
                endcase

                leftsum <= 0;
                rightsum <= 0;

                resetp <= miscreg[2];
                if ((dowrite==0) | (regaddr3!=8'h42)) miscreg[2] <= 0;      // don't accidentally ignore a write

                {egstep, egcount} <= egcount + miscreg[3][4:0];

                end
        regaddr2 <= {2'b10, channel};                       // setup reg address to read waveform/modulation
        clkcount <= clkcount + 1;
        end
3'b001: begin
        feedback <= opvalue[channel];

        regaddr2 <= {2'b11, channel};                       // setup reg address to read envelope parameters
        clkcount <= clkcount + 1;
        end
3'b010: begin
        resetthis <= resetp[channel];

        regaddr2 <= {2'b00, channel};                       // setup reg address to read panning/volume
        clkcount <= clkcount + 1;
        end
3'b011: begin
        regaddr2 <= {2'b01, channel};                       // setup reg address to read octave/freq
        clkcount <= clkcount + 1;
        end
3'b1xx: begin
        // I'll leave this dummy cycle in case I want to add back in register reads later
        oldchan <= channel;
        channel <= channel + 1;
        clkcount <= 0;
        end
endcase


casex (clkcount)                // cycles 2-6: receive and process data pertaining to channel
                                // (cycles 2-4 overlap with register reads for this channel)
                                // (cycles 5-6 overlap with next channel)
3'b010: begin                                               // regread contains waveform/modulation
        waveinvert <= (noisebit & regread[7]) ^ regread[11];
        chanwave <= regread[10:8];
        chanpos <= position[channel];
        egtemp0 <= egvalue[channel];
        chaneg <= egstage[channel];
        chanpm <= regread[3:0];

        if (regread[6:4]==0) feedback2 <= 0;
                else feedback2 <= feedback << regread[6:4];

        chanfms <= ~regread[14:12];
        chanfm <= opvalue[{3'b000, regread[15]}];

        noisecount <= noisecount + 1;
        noisebit <= lfsr[{2'b00, miscreg[0][2:0]}];

//debug1[0] <= noisebit;

        end
3'b011: begin                                               // regread contains envelope config

        if (regread[15:12]==0) egreset <= 1;            // disable EG when attack is 0
                else egreset <= 0;

        egsustain <= regread[11:8];
        if (regread[11:8]==4'hF) egrelease <=1;         // jump immediately to decay2 when sustain is $F
                else egrelease <= 0;

        egtemp <= egtemp0;

        // choose EG rate of change
        casex (chaneg)
        3'b01x: ratetemp <= regread[15:12];             // attack
        3'b1x0: ratetemp <= regread[7:4];               // decay1
        3'b1x1: ratetemp <= regread[3:0];               // decay2
        endcase
        egoverflow <= 0;

        if (chanfms==3'b111) chanfm <= 0;               // disable LFO if sensitivity is at minimum

        // choose phase modulation source data
        case (chanpm[1:0])
        2'b00: phasemod1 <= feedback2;
        2'b01: phasemod1 <= tempvalue2;
        2'b10: phasemod1 <= oldvalue2;
        2'b11: phasemod1 <= oldvalue3;
        endcase

        case (chanpm[3:2])
        2'b00: phasemod2 <= feedback2;
        2'b01: phasemod2 <= tempvalue2;
        2'b10: phasemod2 <= oldvalue2;
        2'b11: phasemod2 <= oldvalue3;
        endcase

        // advance noise shift register when counter reaches target
        if (miscreg[1]==noisecount[18:3]) begin
                noisecount <= 0;
                lfsr <= {lfsr[{2'b00, miscreg[0][6:4]}] ^ noisebit, lfsr[19:1]};
                end

        end
3'b1xx: begin                                               // regread contains panning/volume
        chanpan <= regread[15:6];

        if (egtemp[23:18]==6'h3F) chanvol <= 7'h7F;              // go -95.25dB when note is completely decayed
                else chanvol <= egtemp[23:18] + regread[5:0];                         // uses previous EG value

        waveaddr[7:5] <= chanwave;
        {waveaddr[4:0], highbyte} <= (chanpos[23:9] + phasemod1 + phasemod2) >> 9;        // setup wave address

        casex (chaneg)
        3'b000: egtemp <= egtemp & 24'hFC0000;
        3'b001: {egoverflow, egtemp} <= egtemp + 12'h801;
        3'b01x: {egoverflow, egtemp} <= egtemp - (16 << ratetemp);
        3'b1xx: if (ratetemp!=0) {egoverflow, egtemp} <= egtemp + (1 << ratetemp);
        endcase

        end
3'b000: begin                                               // regread contains octave/freq
        // use oldchan starting now

        chanoct <= regread[15:12];
        chanfreq <= {1'b0, regread[11:0]};

        chanfm2 <= ({{5{chanfm[7]}}, chanfm} * regread[11:7]) >>> 3;

        case (chanvol[2:0])
        3'b000: attenuator <= 256;
        3'b001: attenuator <= 234;
        3'b010: attenuator <= 215;
        3'b011: attenuator <= 197;
        3'b100: attenuator <= 181;
        3'b101: attenuator <= 166;
        3'b110: attenuator <= 152;
        3'b111: attenuator <= 139;
        endcase

        casex ({egoverflow, egtemp[23], egreset})
        3'bxx1: egtemp <= 0;
        3'b100: egtemp[23:16] <= 8'hFC;
        3'b110: egtemp <= 0;
        endcase

        casex ({chaneg, egoverflow, egrelease})
        5'bxxxx1: chaneg <= 5;                                        // key off
        5'b000x0: chaneg <= 1;                                        // DP0 -> DP1
        5'b001x0: if (egtemp[10:0]==11'h7FF) chaneg <= 2;             // DP1 -> attack
        5'b01x10: chaneg <= 4;                                        // attack -> decay1
        5'b1x0x0: if (egtemp[23:20]==egsustain) chaneg <= 5;          // decay1 -> decay2
        endcase

        end
3'b001: begin
        // max range for LFO would be +/- 494 at freq 3968 = +/- 12.45% = +/- 203 cents?
        chanfreq <= ({{9{chanfm2[9]}}, chanfm2} >>> chanfms) + chanfreq;

        tempvalue <= {{8{chansamp[7]}}, chansamp[7:0]} * attenuator;

        opvalue[oldchan] <= chansamp[7:0];

//if (oldchan==2) debug1 <= chaneg;
        end
endcase


casex (clkcount)                // cycles 7-9: finish processing data
                                // (all these cycles overlap with next channel)
3'b010: begin
        if (resetthis) begin
                position[oldchan] <= 0;
                egstage[oldchan] <= 0;
                end else begin
                position[oldchan] <= (chanfreq << chanoct) + chanpos;
                if (egstep) egstage[oldchan] <= chaneg;
                end

        if (egstep) egvalue[oldchan] <= egtemp;

        oldvalue3 <= oldvalue2;
        oldvalue2 <= tempvalue2[15:1];
        tempvalue2 <= tempvalue >>> chanvol[6:3];
        end
3'b011: begin
        chanleft <= {{5{tempvalue2[15]}}, tempvalue2} * chanpan[4:0];
        chanright <= {{5{tempvalue2[15]}}, tempvalue2} * chanpan[9:5];
        end
3'b1xx: begin
        leftsum <= leftsum + {{4{chanleft[20]}}, chanleft[20:5]};
        rightsum <= rightsum + {{4{chanright[20]}}, chanright[20:5]};

        end
endcase


// register reads

waveread <= waverom[waveaddr];

regread <= regfile[regaddr2];


// register writes from outside

regaddr3 <= regaddr;
writedata <= regdatain;
dowrite <= rwrite;

if (dowrite) begin
        case (regaddr3[7:6])
        2'b00: regfile[regaddr3[5:0]] <= writedata;
        2'b01: begin
                miscreg[regaddr3[1:0]] <= writedata;
                if (regaddr3[1:0]==0) lfsr <= 20'hFFFFF;        // init LFSR with all ones when its config
                                                                // register is written
                end
        default: waverom[regaddr3] <= writedata;
        endcase
        end


end



//                not        and        or         xor
// bitwise:       ~          &          |          ^
// logical:       !          &&         ||
//
// logical shifts: << >>
// signed shift right: >>> 
//
// unequal: !=

//parameter linkreg = 4'hF;

//wire trueflag = iword2[11] ? ~(iword2[10] ^ zflag) : ~(iword2[10] ^ cmpflag);

//wire [regmsb:0] predecremented = opdatab2 + {{(regmsb-5){operandc2[5]}}, operandc2};

//        chanfm2 <= chanfm * regread[11:7];                            // does NOT work
//        chanfm <= $signed(chanfm * regread[11:7]) >>> 5;                           // does NOT work
//        chanfm <= $signed({{5{chanfm[10]}}, chanfm} * regread[11:7]) >>> 5;        // works
//        chanfm <= {{5{chanfm[10]}}, chanfm} * regread[11:7] >>> 5;        // works!


endmodule
